import pickle
import warnings
import numpy as np
from tqdm import tqdm
from hdf5storage import loadmat

from ..out_struct import OutStruct

ACCEPTED_CROP_BY = ['onset', 'durations']

def import_outstruct(filepath, strict=True):
    '''
    Import outstruct from matlab (.mat) format. Transpose 'resp' field
    so that it is shape (time, channels).

    Parameters
    ----------
    filepath : string
        Path to .mat file with out structure.
    strict : bool, default=True
        If True, requires strict adherance to the following standards:
        1) Each trial must contain at least the following fields:
        ['name','sound','soundf','resp','dataf']
        2) Each trial must contain the exact same set of fields

    Returns
    -------
    out : naplib.OutStruct object
    
    Notes
    -----
    Given the highly-specific nature of the OutStruct matlab format, this
    function is mostly used internally by Neural Acoustic Processing
    Lab members.
    '''
    loaded = loadmat(filepath)
    loaded = loaded['out'].squeeze()
    fieldnames = loaded[0].dtype.names
    
    req = ['name','sound','soundf','resp','dataf']
    
    
    data = []
    for trial in loaded:
        trial_dict = {}
        for f, t in zip(fieldnames, trial):
            tmp_t = t.squeeze()
            if f == 'resp' or f == 'aud':
                if tmp_t.ndim > 1:
                    tmp_t = tmp_t.transpose(1,0,*[i for i in range(2, tmp_t.ndim)]) # only switch the first 2 dimensions if there are more than 2
            try:
                tmp_t = tmp_t.item()
            except:
                pass
            trial_dict[f] = tmp_t
        data.append(trial_dict)
    
    fieldnames = set(data[0].keys())
    for r in req:
        if strict and r not in fieldnames:
            raise ValueError(f'Missing required field: {r}')
    
    out = OutStruct(data=data, strict=strict)
    return out

def load(filename):
    '''
    Load OutStruct or other object from saved file.
    
    Parameters
    ----------
    filename : string
        File to load. If doesn't end with .pkl this will be added
        automatically.
    
    Returns
    -------
    output : naplib.OutStruct or other object
        Loaded object.
    
    Raises
    ------
    FileNotFoundError
        Can't find file.
    
    '''
    
    if not filename.endswith('.pkl') and '.' not in filename:
        filename = filename + '.pkl'
        
    with open(filename, 'rb') as inp:
        output = pickle.load(inp)

    return output


def save(filename, obj):
    '''
    Save OutStruct or other object with pickle.
    
    Parameters
    ----------
    filename : string
        File to load. If doesn't end with .pkl this will be added
        automatically.
    obj : OutStruct or other object
        Data to save.

    '''
    
    if not filename.endswith('.pkl') and '.' not in filename:
        filename = filename + '.pkl'
    
    with open(filename, 'wb') as f:
        pickle.dump(obj, f)


def read_bids(root,
              subject,
              datatype,
              task,
              suffix,
              session=None,
              befaft=[0, 0],
              crop_by='onset',
              info_include=['sfreq', 'ch_names'],
              resp_channels=None):
    '''
    Read data from the `BIDS file structure <https://bids.neuroimaging.io/>`_ [1]
    to create an OutStruct object. The BIDS file structure is a commonly used structure
    for storing neural recordings such as EEG, MEG, or iEEG.
    
    The channels in the BIDS files are either stored in the 'resp' field of the
    OutStruct or the 'stim' field, depending on whether the `channel_type` is 'stim'.
    
    Please see the :ref:`Importing Data <import data examples>` for more detailed
    tutorials which show how to import external data.
    
    Parameters
    ----------
    root : string, path-like
        Root directory of BIDS file structure.
    datatype : string
        Likely one of ['meg','eeg','ieeg'].
    task : string
        Task name.
    suffix : string
        Suffix name in file naming. This is often the same as datatype.
    session : string
        Session name.
    befaft : list or array-like or length 2, default=[0, 0]
        Amount of time (in sec.) before and after each trial's true duration to include
        in the trial for the OutStruct. For example, if befaft=[1,1] then if each trial's
        recording is 10 seconds long, each trial in the resulting OutStruct will contain
        12 seconds of data, since 1 second of recording before the onset of the event
        and 1 second of data after the end of the event are included on either end.
    crop_by : string, default='onset'
        One of ['onset', 'durations']. If crop by 'onset', each trial is split
        by the onset of each event defined in the BIDS file structure and each
        trial ends when the next trial begins. If crop by 'durations', each trial is split
        by the onset of each event defined in the BIDS file structure and each
        trial lasts the duration specified by the event. This is typically not desired
        when the events are momentary stimulus presentations that have very short duration
        because only the responses during the short duration of the event will be saved, and
        all of the following responses are truncated.
    info_include : list of strings, default=['sfreq, ch_names']
        List of metadata info to include from the raw info. For example, you may wish to include
        other items such as 'file_id', 'line_freq', etc, for later use, if they are stored in
        the BIDS data.
    resp_channels : list, default=None
        List of channel names to select as response channels to be put in the 'resp' field of
        the OutStruct. By default, all channels which are not of type 'stim' will be included.
        Note, the order of these channels may not be conserved.
    
    Returns
    -------
    out : OutStruct
        Event/trial responses, stim, and other basic data in an OutStruct format.
        
    Notes
    -----
    The measurement information that is read-in by this function is stored in the outstruct.mne_info
    attribute. This info can be used in conjunction with
    `mne's visualization functions <https://mne.tools/stable/visualization.html>`_. 
    
    References
    ----------
    .. [1] Pernet, Cyril R., et al. "EEG-BIDS, an extension to the brain imaging
        data structure for electroencephalography." Scientific data 6.1 (2019): 1-5.
    '''
    
    try:
        from mne_bids import BIDSPath, read_raw_bids
    except Exception as e:
        raise Exception('Missing package MNE-BIDS which is required for reading data from BIDS. Please '
            'install it with "pip install --user -U mne-bids" or by following the instructions '
            'at https://mne.tools/mne-bids/stable/install.html')
    
    if crop_by not in ACCEPTED_CROP_BY:
        raise ValueError(f'Invalid "crop_by" input. Expected one of {ACCEPTED_CROP_BY} but got "{crop_by}"')
    
    bids_path = BIDSPath(subject=subject, root=root, session=session, task=task,
                         suffix=suffix, datatype=datatype)
    
    raw = read_raw_bids(bids_path=bids_path, verbose=False)
            
    raws = _crop_raw_bids(raw, crop_by, befaft)
    
    raw_info = None
    
    # figure out which channels are stimulus channels
    stim_channels = [ch for ch, ch_type in zip(raw.ch_names, raw.get_channel_types()) if ch_type == 'stim']

    # for each trial, separate into stim and response channels
    raw_responses = []
    raw_stims = []
    for raw_trial in raws:
        raw_resp = raw_trial.copy().drop_channels(stim_channels)
        if resp_channels is not None:
            raw_resp = raw_resp.pick_channels(resp_channels, verbose=0)
        raw_responses.append(raw_resp)
        
        if raw_info is None:
            raw_info = raw_resp.info

        # if any of the channels are 'stim' channels, store them separately from responses
        if 'stim' in raw_trial.get_channel_types():
            raw_stims.append(raw_trial.pick_types(stim=True, verbose=0))
        else:
            raw_stims.append(None)
    
    # build OutStruct
    outstruct_data = []
    for trial in tqdm(range(len(raws))):
        trial_data = {}
        trial_data['event_index'] = trial
        if 'description' in raw_responses[trial].annotations[0]:
            trial_data['description'] = raw_responses[trial].annotations[0]['description']
        if raw_stims[trial] is not None:
            trial_data['stim'] = raw_stims[trial].get_data().transpose(1,0) # time by channels
            trial_data['stim_ch_names'] = raw_stims[trial].info['ch_names']
        trial_data['resp'] = raw_responses[trial].get_data().transpose(1,0) # time by channels
        trial_data['befaft'] = befaft
        for info_key in info_include:
            if info_key not in info_include:
                warnings.warn(f'info_include key "{info_key}" not found in raw info')
            else:
                trial_data[info_key] = raw_responses[trial].info[info_key]
        outstruct_data.append(trial_data)  

    outstruct = OutStruct(outstruct_data, strict=False)
    outstruct.set_mne_info(raw_info)
    return outstruct
    
    
def _crop_raw_bids(raw_instance, crop_by, befaft):
    '''
    Crop the raw data to trials based on events in its annotations.
    
    Parameters
    ----------
    raw_instance : mne.io.Raw-like object
    
    crop_by : string, default='onset'
        One of ['onset', 'annotations']. If crop by 'onset', each trial is split
        by the onset of each event defined in the BIDS file structure and each
        trial ends when the next trial begins. If crop by 'annotations', each trial is split
        by the onset of each event defined in the BIDS file structure and each
        trial lasts the duration specified by the event. This is typically not desired
        when the events are momentary stimulus presentations that have very short duration
        because only the responses during the short duration of the event will be saved, and
        all of the following responses are truncated.
    
     Returns
     -------
     raws : list
         The cropped raw objects.

    '''

    max_time = (raw_instance.n_times - 1) / raw_instance.info['sfreq']
    
    raws = []
    for i, annot in enumerate(raw_instance.annotations):
        onset = annot["onset"] - raw_instance.first_time - befaft[0]
        if -raw_instance.info['sfreq'] / 2 < onset < 0:
            onset = 0
        if crop_by == 'onset':
            if i == len(raw_instance.annotations)-1:
                tmax = max_time
            else:
                if befaft[1] > 0:
                    warnings.warn('befaft[1] is positive, but crop_by is "onset", so the ending of each trial will include a portion of the next trial')
                tmax = raw_instance.annotations[i+1]["onset"] + befaft[1]
            tmax = min([tmax, max_time])
            raw_crop = raw_instance.copy().crop(onset, tmax)
        
        else:
            tmax = onset + annot["duration"] + befaft[1]
            tmax = min([tmax, max_time])
            raw_crop = raw_instance.copy().crop(onset, tmax)
        
        raws.append(raw_crop)
    
    return raws
    



"""
Global variables and helpers for forced alignment
"""

import bisect
import logging
import os
import yaml

# global variables

SP = "sp"
SIL = "sil"
TEMP = "temp"

EPOCHS = 5

MISSING = "missing.txt"
OOV = "OOV.txt"

CONFIG = "config.yaml"
DICT = "dict"
HMMDEFS = "hmmdefs"
MACROS = "macros"
PROTO = "proto"
VFLOORS = "vFloors"

ALIGNED = ".aligned.mlf"
SCORES = ".scores.csv"


# samplerates which appear to be HTK-compatible (all divisors of 1e7)
SAMPLERATES = [4000, 8000, 10000, 12500, 15625, 16000, 20000, 25000,
               31250, 40000, 50000, 62500, 78125, 80000, 100000, 125000,
               156250, 200000]


# helpers

def opts2cfg(filename, opts):
    """
    Convert dictionary of key-value pairs to an HTK config file
    """
    with open(filename, "w") as sink:
        for (setting, value) in opts.items():
            print("{!s} = {!s}".format(setting, value), file=sink)


def mkdir_p(dirname):
    """
    Create a directory, recursively if necessary, and suceed
    silently if it already exists
    """
    os.makedirs(dirname, exist_ok=True)


def splitname(fullname):
    """
    Split a filename into directory, basename, and extension
    """
    (dirname, filename) = os.path.split(fullname)
    (basename, ext) = os.path.splitext(filename)
    return (dirname, basename, ext)


def resolve_opts(args):
    if args.configuration is None:
        logging.error("Configuration (-c) file not specified.")
        exit(1)
    with open(args.configuration, "r") as source:
        try:
            opts = yaml.load(source, Loader=yaml.FullLoader)
        except yaml.YAMLError as err:
            logging.error("Error in configuration file: %s", err)
            exit(1)
    # command line only
    if not args.dictionary:
        logging.error("Dictionary (-d) not specified.")
        exit(1)
    opts["dictionary"] = args.dictionary
    if not args.epochs:
        args.epochs = EPOCHS
    opts["epochs"] = args.epochs
    # could be either, and the command line takes precedent.
    try:
        sr = args.samplerate if args.samplerate else opts["samplerate"]
    except KeyError:
        logging.error("Samplerate (-s) not specified.")
        exit(1)
    if sr not in SAMPLERATES:
        i = bisect.bisect(SAMPLERATES, sr)
        if i == 0:
            pass
        elif i == len(SAMPLERATES):
            i = -1
        elif SAMPLERATES[i] - sr > sr - SAMPLERATES[i - 1]:
            i = i - 1
        # else keep `i` as is
        sr = SAMPLERATES[i]
        logging.warning("Using {} Hz as samplerate".format(sr))
    opts["samplerate"] = sr
    return opts



# Copyright (c) 2011-2014 Kyle Gorman and Michael Wagner
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""
Command-line driver for the module
"""

import logging
import os
import sys
import yaml

from bisect import bisect
from shutil import copyfile
from textgrid import MLF

from corpus import Corpus
from aligner import Aligner
from archive import Archive
from utilities import splitname, resolve_opts, \
                       ALIGNED, CONFIG, HMMDEFS, MACROS, SCORES

from argparse import ArgumentParser

DICTIONARY = "eng.dict"
MODEL = "eng.zip"

LOGGING_FMT = "%(message)s"


# parse arguments
argparser = ArgumentParser(prog="{} -m aligner".format(sys.executable),
                           description="Prosodylab-Aligner")
argparser.add_argument("-c", "--configuration",
                       help="config file")
argparser.add_argument("-d", "--dictionary", metavar="DICT", action="append",
                       help="dictionary file (default: {}) (can specify multiple)".format(DICTIONARY))
argparser.add_argument("-s", "--samplerate", type=int,
                       help="analysis samplerate (in Hz)")
argparser.add_argument("-e", "--epochs", type=int,
                       help="# of epochs of training per round")
input_group = argparser.add_argument_group()
input_group.add_argument("-r", "--read",
                         help="source for a precomputed acoustic model")
input_group.add_argument("-t", "--train",
                         help="directory containing data for training")
output_group = argparser.add_mutually_exclusive_group(required=True)
output_group.add_argument("-a", "--align",
                          help="directory containing data to align")
output_group.add_argument("-w", "--write",
                          help="destination for computed acoustic model")
verbosity_group = argparser.add_mutually_exclusive_group()
verbosity_group.add_argument("-v", "--verbose", action="store_true",
                             help="Verbose output")
verbosity_group.add_argument("-V", "--extra-verbose", action="store_true",
                             help="Even more verbose output")
args = argparser.parse_args()

# hack to allow proper override of default dictionary
if not args.dictionary:
    args.dictionary = [DICTIONARY]

# set up logging
loglevel = logging.WARNING
if args.extra_verbose:
    loglevel = logging.DEBUG
elif args.verbose:
    loglevel = logging.INFO
logging.basicConfig(format=LOGGING_FMT, level=loglevel)

# input: pick one
if args.train:
    if args.read:
        logging.error("Cannot train on persistent model.")
        exit(1)
    logging.info("Preparing corpus '{}'.".format(args.train))
    opts = resolve_opts(args)
    corpus = Corpus(args.train, opts)
    logging.info("Preparing aligner.")
    aligner = Aligner(opts)
    logging.info("Training aligner on corpus '{}'.".format(args.train))
    aligner.HTKbook_training_regime(corpus, opts["epochs"],
                                    flatstart=(args.read is None))
else:
    if not args.read:
        args.read = MODEL
    logging.info("Reading aligner from '{}'.".format(args.read))
    # warn about irrelevant flags
    if args.configuration:
        logging.warning("Ignoring config flag (-c/--configuration).")
        args.configuration = None
    if args.epochs:
        logging.warning("Ignoring epochs flag (-e/--epochs).")
    if args.samplerate:
        logging.warning("Ignoring samplerate flag (-s/--samplerate).")
        args.samplerate = None
    # create archive from -r argument
    archive = Archive(args.read)
    # read configuration file therefrom, and resolve options with it
    args.configuration = os.path.join(archive.dirname, CONFIG)
    opts = resolve_opts(args)
    # initialize aligner and set it to point to the archive data
    aligner = Aligner(opts)
    aligner.curdir = archive.dirname

# output: pick one
if args.align:
    # check to make sure we're not aligning on the training data
    if (not args.train) or (os.path.realpath(args.train) !=
                            os.path.realpath(args.align)):
        logging.info("Preparing corpus '{}'.".format(args.align))
        corpus = Corpus(args.align, opts)
    logging.info("Aligning corpus '{}'.".format(args.align))
    aligned = os.path.join(args.align, ALIGNED)
    scores = os.path.join(args.align, SCORES)
    aligner.align_and_score(corpus, aligned, scores)
    logging.debug("Wrote MLF file to '{}'.".format(aligned))
    logging.debug("Wrote likelihood scores to '{}'.".format(scores))
    logging.info("Writing TextGrids.")
    size = MLF(aligned).write(args.align)
    if not size:
        logging.error("No paths found!")
        exit(1)
    logging.debug("Wrote {} TextGrids.".format(size))
elif args.write:
    # create and populate archive
    (_, basename, _) = splitname(args.write)
    archive = Archive.empty(basename)
    archive.add(os.path.join(aligner.curdir, HMMDEFS))
    archive.add(os.path.join(aligner.curdir, MACROS))
    # whatever this is, it's not going to work once you move the data
    if "dictionary" in opts:
        del opts["dictionary"]
    with open(os.path.join(archive.dirname, CONFIG), "w") as sink:
        yaml.dump(opts, sink)
    (basename, _) = os.path.splitext(args.write)
    archive_path = os.path.relpath(archive.dump(basename))
    logging.info("Wrote aligner to '{}'.".format(archive_path))
# else unreachable

logging.info("Success!")



# Copyright (c) 2011-2014 Kyle Gorman and Michael Wagner
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""
Utilities for audio resampling (etc.)
"""

import wave

from numpy import asarray
from scipy.io import wavfile
from scipy.signal import resample


class WavFile(object):

    """
    Class representing a mono wav file
    """

    def __init__(self, signal, Fs):
        self.signal = asarray(signal)
        self.Fs = Fs

    @staticmethod
    def samplerate(filename):
        """
        Get samplerate without reading the entire wav file into memory
        """
        with wave.open(filename, "r") as source:
            return source.getframerate()

    @classmethod
    def from_file(cls, filename):
        (Fs, signal) = wavfile.read(filename)
        if signal.ndim > 1:
            raise ValueError("Expected mono audio," +
                             " but '{}'".format(filename) +
                             " has {} channels.".format(signal.ndim))
        return cls(signal, Fs)

    def __repr__(self):
        return "{}(signal={!r}, Fs={!r})".format(self.__class__.__name__,
                                                 self.signal, self.Fs)

    def __len__(self):
        return len(self.signal)

    def write(self, filename):
        wavfile.write(filename, self.Fs, self.signal)

    def _resample(self, Fs_out):
        ratio = Fs_out / self.Fs
        resampled_signal = resample(self.signal, int(ratio * len(self)))
        return resampled_signal

    def resample(self, Fs_out):
        return WavFile(self._resample(Fs_out), Fs_out)

    def resample_bang(self, Fs_out):
        self.signal = self._resample(Fs_out)
        self.Fs = Fs_out



# Copyright (c) 2011-2014 Kyle Gorman and Michael Wagner
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""
Aligner utilities
"""

import os
import logging

from re import match
from tempfile import mkdtemp
from shutil import copyfile, rmtree
from subprocess import check_call, Popen, CalledProcessError, PIPE

from utilities import opts2cfg, mkdir_p, \
                       HMMDEFS, MACROS, PROTO, SP, SIL, TEMP, VFLOORS


# regexp for parsing the HVite trace
HVITE_SCORE = r".+==  \[\d+ frames\] (-\d+\.\d+)"
# in case you"re curious, the rest of the trace string is:
#     /\[Ac=-\d+\.\d+ LM=0.0\] \(Act=\d+\.\d+\)/


class Aligner(object):

    """
    Class representing an aligner, including HMM definitions and 
    configuration options
    """

    def __init__(self, opts):
        # make temporary directories to stash everything
        hmmdir = os.environ["TMPDIR"] if "TMPDIR" in os.environ else None
        self.hmmdir = mkdtemp(dir=hmmdir)
        # config options
        self.HCompV_opts = opts["HCompV"]
        self.HERest_cfg = os.path.join(self.hmmdir, "HERest.cfg")
        opts2cfg(self.HERest_cfg, opts["HERest"])
        self.HVite_opts = opts["HVite"]
        self.pruning = [str(i) for i in opts["pruning"]]
        # initialize directories
        self.epochs = 0
        self.curdir = os.path.join(self.hmmdir, str(self.epochs).zfill(3))
        mkdir_p(self.curdir)
        self.epochs += 1
        self.nxtdir = os.path.join(self.hmmdir, str(self.epochs).zfill(3))
        mkdir_p(self.nxtdir)

    def _nxtdir(self):
        """
        Get the next HMM directory
        """
        self.curdir = self.nxtdir
        self.epochs += 1
        self.nxtdir = os.path.join(self.hmmdir, str(self.epochs).zfill(3))
        mkdir_p(self.nxtdir)

    def flatstart(self, corpus):
        self.epochs = 1
        # make `proto`
        self.proto = os.path.join(self.hmmdir, PROTO)
        with open(self.proto, "w") as proto:
            # FIXME this is highly specific to the default acoustic
            # features, but figuring out the number of means and variances
            # needed from the HCopy configuration file is not trivial.
            means = " ".join(["0.0" for _ in range(39)])
            varg = " ".join(["1.0" for _ in range(39)])
            print("""~o <VECSIZE> 39 <MFCC_D_A_0>
~h "proto"
<BEGINHMM>
<NUMSTATES> 5""", file=proto)
            for i in range(2, 5):
                print("<STATE> {}".format(i), file=proto)
                print("<MEAN> 39", file=proto)
                print(means, file=proto)
                print("<VARIANCE> 39", file=proto)
                print(varg, file=proto)
            print("""<TRANSP> 5
 0.0 1.0 0.0 0.0 0.0
 0.0 0.6 0.4 0.0 0.0
 0.0 0.0 0.6 0.4 0.0
 0.0 0.0 0.0 0.7 0.3
 0.0 0.0 0.0 0.0 0.0
<ENDHMM>""", file=proto)
        # make `vFloors`
        check_call(["HCompV", "-m",
                              "-f", str(self.HCompV_opts["F"]),
                              "-C", self.HERest_cfg,
                              "-S", corpus.feature_scp,
                              "-M", self.curdir, self.proto])
        # make `macros`
        # get first three lines from local proto
        with open(os.path.join(self.curdir, MACROS), "a") as macros:
            with open(os.path.join(self.curdir,
                      os.path.split(self.proto)[1]), "r") as proto:
                for _ in range(3):
                    print(proto.readline().strip(), file=macros)
            # get remaining lines from `vFloors`
            with open(os.path.join(self.curdir, VFLOORS), "r") as vfloors:
                print("".join(vfloors.readlines()).rstrip(), file=macros)
        # make `hmmdefs`
        with open(os.path.join(self.curdir, HMMDEFS), "w") as hmmdefs:
            with open(os.path.join(self.curdir, PROTO), "r") as proto:
                protolines = proto.readlines()[4:]
            with open(corpus.phons, "r") as phons:
                for phone in phons:
                    print('~h "{}"'.format(phone.rstrip()), file=hmmdefs)
                    print("".join(protolines).rstrip(), file=hmmdefs)

    def train(self, corpus, epochs):
        """
        Perform one or more rounds of estimation
        """
        for _ in range(epochs):
            logging.debug("Training iteration {}.".format(self.epochs))
            check_call(["HERest", "-C", self.HERest_cfg,
                        "-S", corpus.feature_scp,
                        "-I", corpus.phon_mlf,
                        "-M", self.nxtdir,
                        "-H", os.path.join(self.curdir, MACROS),
                        "-H", os.path.join(self.curdir, HMMDEFS),
                        "-t"] + self.pruning + [corpus.phons])
            self._nxtdir()

    def small_pause(self, corpus):
        """
        Add in a tied-state small pause model
        """
        # make new hmmdef
        saved = ['~h "{}"'.format(SP)]
        # opened both for reading and writing
        with open(os.path.join(self.curdir, HMMDEFS), "r+") as hmmdefs:
            # find SIL
            for line in hmmdefs:
                if line.startswith('~h "{}"'.format(SIL)):
                    break
            saved.extend(["<BEGINHMM>",
                          "<NUMSTATES> 3",
                          "<STATE> 2"])
            # pass until we get to SIL's middle state
            for line in hmmdefs:
                if line.startswith("<STATE> 3"):
                    break
            # grab SIL's middle state
            for line in hmmdefs:
                if line.startswith("<STATE> 4"):
                    break
                saved.append(line.rstrip())
            # add in the TRANSP matrix
            saved.extend(["<TRANSP> 3",
                          " 0.0 1.0 0.0",
                          " 0.0 0.9 0.1",
                          " 0.0 0.0 0.0",
                          "<ENDHMM>"])
            # write all the lines to the end of `hmmdefs`
            hmmdefs.seek(0, os.SEEK_END)
            hmmdefs.write("\n".join(saved))
        # tie states together
        temp = os.path.join(self.hmmdir, TEMP)
        with open(temp, "w") as hed:
            print("""AT 2 4 0.2 {{{1}.transP}}
AT 4 2 0.2 {{{1}.transP}}
AT 1 3 0.3 {{{0}.transP}}
TI silst {{{1}.state[3],{0}.state[2]}}
""".format(SP, SIL), file=hed)
        check_call(["HHEd", "-H", os.path.join(self.curdir, MACROS),
                            "-H", os.path.join(self.curdir, HMMDEFS),
                            "-M", self.nxtdir,
                            temp, corpus.phons])
        temp = os.path.join(self.hmmdir, TEMP)
        with open(temp, "w") as led:
            print("""EX
IS {0} {0}
""".format(SIL), file=led)
        check_call(["HLEd", "-l", corpus.auddir,
                            "-d", corpus.taskdict,
                            "-i", corpus.phon_mlf,
                            temp, corpus.word_mlf])
        logging.debug("(Skipping an iteration number).")
        self._nxtdir()

    def align(self, corpus, mlf):
        check_call(["HVite", "-a", "-m",
                             "-o", "SM",
                             "-y", "lab",
                             "-b", SIL,
                             "-i", mlf,
                             "-L", corpus.labdir,
                             "-C", self.HERest_cfg,
                             "-S", corpus.feature_scp,
                             "-H", os.path.join(self.curdir, MACROS),
                             "-H", os.path.join(self.curdir, HMMDEFS),
                             "-I", corpus.word_mlf,] +
                             #FIXME(kg) do we want this?
                             #"-s", str(self.HVite_opts["SFAC"]), 
                             #"-t"] + self.pruning +
                   [corpus.taskdict, corpus.phons], stdout=PIPE)

    def realign(self, corpus):
        """
        Align and then overwrite `corpus.word_mlf` with the result
        """
        temp = os.path.join(self.hmmdir, TEMP)
        self.align(corpus, temp)
        copyfile(temp, corpus.word_mlf)

    def align_and_score(self, corpus, mlf, scores):
        """
        The same as `self.align`, but also generates a text file `score`
        with -log likelihood confidence scores for each audio file
        """
        proc = Popen(["HVite", "-a", "-m",
                               "-T", "1",
                               "-o", "SM",
                               "-y", "lab",
                               "-b", SIL,
                               "-i", mlf,
                               "-L", corpus.labdir,
                               "-C", self.HERest_cfg,
                               "-S", corpus.feature_scp,
                               "-H", os.path.join(self.curdir, MACROS),
                               "-H", os.path.join(self.curdir, HMMDEFS),
                               "-I", corpus.word_mlf,] +
                               #FIXME(kg) do we want this?
                               #"-s", str(self.HVite_opts["SFAC"]),
                               #"-t"] + self.pruning +
                     [corpus.taskdict, corpus.phons],
                     stdout=PIPE)
        with open(scores, "w") as sink:
            i = 0
            for line in proc.stdout:
                m = match(HVITE_SCORE, line.decode("UTF-8"))
                if m:
                    print('"{!s}",{!s}'.format(corpus.audiofiles[i],
                                               m.group(1)), file=sink)
                    i += 1
        # Popen equivalent to check_call...
        retcode = proc.wait()
        if retcode != 0:
            raise CalledProcessError(retcode, proc.args)

    def HTKbook_training_regime(self, corpus, epochs, flatstart=True):
        if flatstart:
            logging.info("Flat start training.")
            self.flatstart(corpus)
        self.train(corpus, epochs)
        logging.info("Modeling silence.")
        self.small_pause(corpus)
        logging.info("Additional training.")
        self.train(corpus, epochs)
        logging.info("Realigning.")
        self.realign(corpus)
        logging.info("Final training.")
        self.train(corpus, epochs)

    def __del__(self):
        rmtree(self.hmmdir)
